/* JavaGame is a game engine used to teach computer programming.
 * There is a corresponding JavaScript version to use in a web page.
 *
 * GameWgt is the base class for graphic elements in JavaGame.
 * It includes a (static) graphics engine for drawing on a pixel buffer.
 *
 * JavaGame copyright 2020 Itty Bitty Computers and released at this time
 * to the public as open source. There are no warranties of any kind.
 */

/**
 * Base class for game board widgets.                  // 2021 July 15
 */
public class GameWgt   {

 /**
  * These constants are used internally in this base class. They are copied
  * from class JavaGame. If you need them, you can get them from there.
  */
  private final static boolean Logging = JavaGame. Logging,
      GoodOptn = JavaGame. GoodOptn,
      GoodDiag = JavaGame. GoodDiag;
  public final static boolean GoodTxt = false;
  private final static int Max_Log = JavaGame. Max_Log;
  private static int WinHi = JavaGame. Win_Hi,
      WinWi = JavaGame. Win_Wi,
      MaxWinSize = WinHi*0x10000+WinWi,
      FustFocLog = JavaGame. Fust_Log,
      MaxLogWgts = Max_Log, MaxLogAnim = FustFocLog>>1,
      SeenLog = -1; // bits log each milestone in starting up

  /**
  * The instance data used in this class.
  *
  * Gtype is the type of widget in the low 8 bits, and flag bits above that:
  *   =0: (GameWgt) plain rectangle      +256: Collidable ("Bump")
  *   =1: pixel image (window image)     +512: Type-able ("Key")
  *   =2: (GameIcon) 1-color icon       +1024: Clickable
  *   =3: (GameFont) font definition    +2048: Dragable
  *   =4: (GameTxLn) text line          +4096: Rollover-able
  *   =5: group                         +8192: no-focus (GameScro only)
  *   =6: (GameSprit) sprite group
  *   =7: (GameScro) scroll group      +32768: off edge (set by Collider)
  *
  * Gcolor is the color to draw this widget
  * Gtop, Gleft are the vertical and horizontal coordinates of this widget
  *   relative to Parent's position, or absolute if Parent=0
  * Gtall, Gwide are the dimensions of this widget's containing rectangle
  *
  * Ginfo is additional information for this widget:
  *   (type=1) ref+ of the first widget in the game
  *   (type=2) animation info, bit fields, high to low:
  *      (frame rate)*16:6, (+ frames)-1:6, (step size):6, (pixel offset):14
  *   (type=3) character offset current selected character
  *   (type=4) (cursor H-posn):12, (cursor height):5, (ref+ of the font):14
  *   (type=6) packed velocity, (vertical*256):16, (horizontal*256):16
  *   (type=7) packed scroll position
  *    (other) may be packed 4-char identifier for this widget or its group
  *
  * Parent is the ref+ of the widget (type=5/6/7) listing this widget as child
  * RefNum is the ref+ of this widget = its index in AllWgts[]
  *
  * Gdata is an array of additional data (but not if type=0)
  *   (type=1) the pixel array for this image, each int = RGB for one pixel
  *   (type=2) the pixels for this icon, each int = 32 vertical pixels
  *   (type=3) the pixels for this font, each int = 32 vertical pixels
  *   (type=4) the text of this line, each int packed 4 chars
  *   (type=5) ref+s of each child widget
  *   (type=6) ref+s of each child widget (rect or icon or text)
  *   (type=7) ref+s of group being scrolled, then V-scrollbar, then H-scrollbar
  *
  * Gselim (GameTxLn/GameScro only) = current selection or scroll limits
  */
  int Gtype, Gcolor, Gtop, Gleft, Gtall, Gwide, Ginfo, Parent, RefNum;
  int[] Gdata; // ty=1: image pix, 2: 32-row pix, 3: index+pix,
    //   4: packed text, 5,6,7: array of RefNum's
  int Gselim; // = current selection (GameTxLn) or scroll limits (GameScro)

  // Graphics Engine Data..

 /**
  * All GameEngine drawing is done in this class.
  *
  * SceneTall,SceneWide are the image dimensions, set by SetMyScreen
  * theScreen is the array of RGB pixels [PixScale is not used]
  */
  private static  int SceneTall = 0, SceneWide = 0, PixScale = 0;
  private static int[] theScreen = null ;

  // Graphics Engine Public Methods..

 /**
  * Sets the pixel at the specified row and column to a given color.
  *
  * @param colo   The color to go in that position, = 0x00RRGGBB
  * @param rx     The pixel row
  * @param cx     The pixel column
  */
  public static void PokePixel(int colo, int rx, int cx)  {
    int[] myPix = theScreen; // -> theScreen
    int here = SceneWide ;
    here = rx*SceneWide+cx;
    if (myPix != null) if (here < myPix.length)
      if (here >= 0) myPix[here] = colo;} //~PokePixel

 /**
  * Sets a whole rectangle of pixels from one corner of the
  * rectangle in pixel coordinates, to the opposite corner.
  *
  * To draw a horizontal line, use the same value for rx=rz;
  * to draw a vertical line, use the same value for cx=cz.
  *
  * @param colo   The pixel color to fill that rectangle, = 0x00RRGGBB
  * @param rx     The pixel row in one corner
  * @param cx     The pixel column in the same corner
  * @param rz     The pixel row in the other corner (inclusive)
  * @param cz     The pixel column there
  */
  public static void RectFill(int colo, int rx, int cx, int rz, int cz)  {
    int here, thar, whar, ix;   // top,  left,  bottom,  right // BR not sz
    int[] myPix = theScreen;
    if (myPix == null) return;
    if (rz<rx)  {here = rz; rz = rx; rx = here;}
    if (cz<cx)  {here = cz; cz = cx; cx = here;}
    if (rx<0) rx = 0;
    if (rz >= SceneTall) rz = SceneTall-1;
    if (cx<0) cx = 0;
    if (cz >= SceneWide) cz = SceneWide-1;
    whar = rx*SceneWide+cx;
    for (rx=rx; rx <= rz; rx++) {
      thar = whar;
      whar = whar+SceneWide;
      for (here=cx; here <= cz; here++) {
        if (myPix == null) break;
        if (thar >= 0) if (thar < myPix.length)
          myPix[thar] = colo;
        thar++;}}} //~RectFill

 /**
  * Sets the graphics pixel buffer to the client's array.
  *
  * @param theAry The pixel array, at least tall*wide in length
  * @param tall   The number of pixel rows
  * @param wide   The number of pixel columns
  */
  public static void SetMyScreen(int[] theAry, int tall, int wide)  {
    int size = tall*wide;
    if (theAry == null) return;
    if (size > theAry.length) return;
    if (tall <= 0) return;
    if (wide <= 0) return;
    theScreen = theAry;
    SceneTall = tall;
    SceneWide = wide;
    WinHi = tall;
    WinWi = wide;
    MaxWinSize = tall*0x10000+wide;
    PixScale = 0;} //~SetMyScreen

  // Handy (private/utility) methods..

  private int GetMilliSecs() {return JavaGame.GetMilliSecs();}

  private GameWgt GetIxWgt(int ix) {return JavaGame.GetIxWgt(ix);}

/**
  * Internal utility for passing (behind the scenes) a value
  * that can be logged as a unique callerID
  *
  * @param  whom   The number to be passed, or =0 to clear it
  *
  * @return        Its previous value
  */
  public static int WhoCalled(int whom)  { // reads out & sets CallersID
    int zx = JavaGame. CallersID;
    JavaGame. CallersID = whom;
    return zx;} //~WhoCalled

 /**
  * A safe (no exceptions) non-OOPS way to extract a character from a string.
  *
  * @param  here  The (0-base) character position
  * @param  aStr  The string from which to extract that character
  *
  * @return       The character extracted, or '\0' if out of bounds
  */
  public static char CharAt(int here, String aStr) { // Java version..
    if (here<0) return '\0';
    if (aStr == null) return '\0';
    try {
      if (here >= aStr.length()) return '\0';
      return aStr.charAt(here);}
    catch (Exception ex) {return '\0';}} //~CharAt

 /**
  * A safe (no exceptions) non-OOPS way to get the length of a string.
  *
  * @param  aStr  The string to get the length of
  *
  * @return       Its length, or 0 if null
  */
  public static int StrLength(String aStr) { // Java version..
    if (aStr == null) return 0;
    return aStr.length();} //~StrLength

/**
  * Strongly-typed (not overloaded) string selector.
  * This is part of a collection of tools to build debug print lines.
  *
  * @param  whom  The selector
  * @param  tru   Returned if whom=true
  * @param  fls   Returned if whom=false
  *
  * @return       The selected string, tru if whom=true, else fls
  */
  private String IffyStr(boolean whom, String tru, String fls)  {
    if (whom) return tru;
    return fls;} //~IffyStr

 /**
  * A safe (no exceptions) way to get the rest of a string.
  * Text past the end is considered empty.
  *
  * @param  here  The (0-base) position to start
  * @param  aStr  The string from which to extract that substring
  *
  * @return       The rest of the string, after here
  */
  private String RestOf(int here, String aStr)  {
    int lxx;
    try {
    if (here<0) return "";
    lxx = StrLength(aStr);
    if (lxx==0) return "";
    if (here >= lxx) return "";
    return aStr.substring(here);
    } catch (Exception ex)  {}
    return "";  } //~RestOf

  /**
  * A safe (no exceptions) way to extract a substring.
  * Text past the end is empty.
  *
  * @param  here  The (0-base) position to start
  * @param  lxx   The desired length (the result could be less)
  * @param  aStr  The string from which to extract that substring
  *
  * @return       The substring
  */
  private String Substring(int here, int lxx, String aStr)  {
    int nx = StrLength(aStr);
    try {
    if (here<0) return "";
    if (lxx <= 0) return "";
    if (nx==0) return "";
    if (here >= nx) return "";
    lxx = here+lxx;
    if (lxx >= nx) return RestOf(here,aStr);
    return aStr.substring(here,lxx);
    } catch (Exception ex)  {}
    return "";  } //~Substring

  /**
  * Formats an integer as hexadecimal and returns it with a prefix and suffix.
  * This is useful when printing numbers that you want to see the bits of.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  * @param  after  The suffix
  *
  * @return       The combined string
  */
  private String SeeHex(String before, int whom, String after)  {
    int shft = 28, lxx = StrLength(before); // special-case URL encoding..
    if (lxx==1) if ((whom&-256)==0) if (StrLength(after)==0)
        if ((int)(CharAt(0,before)) == 37)  { // '%'
      if (whom>15)  {
        before = before + SeeHex("",whom>>4,"");
        whom = whom&15;} //~if
      else before = before + "0";} //~if
    if (whom<0)  {
      before = SeeHex(before,(whom>>28)&15,"");
      if (whom+0x00100000 >= 0)  { // remove leading 'F's..
        shft = 16;
        while ((1<<shft)+whom >= 0)  {
          shft = shft-4;
          if (shft==0) break;} //~while
        while (shft>0)  { // skip over (internal) leading zeros..
          if (((15<<shft)&whom) !=0) break;
          shft = shft-4;} //~while
        whom = whom&((16<<shft)-1);
        before = before + "-F";} //~if
      else whom = whom&0x0FFFFFFF;} //~if // (whom<0)
    if (whom>15) before = SeeHex(before,whom>>4,"");
    return before + Substring(whom&15,1,"0123456789ABCDEF")
        + after;} //~SeeHex

  /**
  * Extracts and sign-extends the second of a pair of half-integers.
  *
  * @param  here  The pair
  *
  * @return       The low half, sign-extended to a full int
  */
  private int SignExtend(int here)  { // low 16 bits signed
    return (here&0x7FFF)-(here&0x8000);} //~SignExtend

  /**
  * Formats a 32-bit integer as two 16-bit values and returns it with a prefix
  * and suffix. This is useful when printing numbers that pack two smaller
  * integers (like board coordinates) into a single number.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  * @param  after  The suffix
  *
  * @return       The combined string
  */
  private String LogVH(String before, int whom, String after)  {
    return before + (whom>>16) + "/" + SignExtend(whom) + after;} //~LogVH

/**
  * Adds two pairs of half-integers in a single int.
  *
  * @param  here  The first pair
  * @param  thar  The second pair
  *
  * @return       The pair of sums
  */
  private int AddPair(int here, int thar)  {
    return (here&-0x10000)+(thar&-0x10000)
        + ((here+thar)&0xFFFF);} //~AddPair

  /**
  * Returns in a single int the negatives of a pair of half-integers.
  *
  * @param  here  The pair
  *
  * @return       Its negative
  */
  private int PairNeg(int here)  {
    return ((-here)&0xFFFF)-(here&-0x10000);} //~PairNeg

/**
  * A safe (no exceptions) way to get the size of a (possibly null) array.
  *
  * @param  data  The array
  *
  * @return       Its length, or =0 if null
  */
  private int SafeArySz(int[] data)  {
    if (data == null) return 0;
    return data.length;} //~SafeArySz

  /**
  * A safe (no exceptions) way to get a (possibly null) array element.
  * Returns a specified default value if the array is null or the given
  * index is out of bounds.
  *
  * @param  ix      The index into the array
  * @param  data    The array
  * @param  defalt  A default value to return if can't access
  *
  * @return       Its length, or =0 if null
  */
  private int SafeAryElt(int ix, int[] data, int defalt)  {
    while (true)  {
      if (data == null) break;
      if (ix<0) break;
      if (ix >= data.length) break;
      defalt = data[ix];
      break;} //~while
    return defalt;} //~SafeAryElt

  /**
  * Formats a 16-bit integer as 8.8 fixed-point and returns it with a prefix.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  *
  * @return       The result string
  */
  private String FormFixt(String before, int whom)  {
    int frac = whom&255, zx = frac*10+6;   // assumes 8-bit fraction
    String info = "";                     // ..and signed 8+bit integer part
    boolean shorten = ((whom>25599)||(StrLength(before)==1));
    if (false) if ((whom&-0x8000) !=0) if (whom+0x10000 != (whom&0xFFFF))
      whom = SignExtend(whom); // caller did
    if (whom<0) return FormFixt(before + "-",-whom);
    before = before + (whom>>8);
    if (frac>0)  {
      before = before + ".";
      if (frac==128) info = "5";
      else if (frac==64) info = "25";
      else if (frac==192) info = "75";
      else if (frac==1)  {
        if (shorten) info = "01"; else
          info = "004";} //~if
      else if ((zx&255)<13) info = info + (zx>>8); // rounds to 1 digit
      else if (shorten)  { // round to 2 fractional digits..
        frac = (frac*100+127)>>8;
        if (frac<10) info = "0";
        else if (frac>99) frac = 99;
        info = info + frac;} //~if
      else if ((frac&31) !=0)  { // round to 3 fractional digits..
        frac = (frac*1000+127)>>8;
        if (frac<10) info = "00";
        else if (frac<100) info = "0";
        else if (frac>999) frac = 999;
        info = info + frac;} //~if
      else info = info + ((frac>>5)*125);} //~if
    return before + info;} //~FormFixt

  /**
  * Formats an integer as hexadecimal if larger than 32K and returns it
  * with a prefix and suffix, for easier reading.
  *
  * @param  before The prefix
  * @param  whom   The number to be formatted
  * @param  after  The suffix
  *
  * @return       The result string
  */
  private String HexIfMore(String before, int whom, String after)  {
    if (SignExtend(whom)==whom) return before + whom + after;
    return SeeHex(before + "0x",whom,after); } //~HexIfMore

/**
  * A safe (no exceptions) way to extract an item (word/line) from a string.
  * When getting words, excess white space is ignored, so the returned
  * string is empty only when requesting a word beyond the last. Otherwise
  * consecutive delimiters result in empty items being returned, so that
  * you always get the nth item of a multi-item string. Items past the end
  * or before the front (0 or less) are empty.
  *
  * @param  delim  Use '\n' to get a whole line, ' ' to get a word
  * @param  whom   The (1-base) item number to get
  * @param  aStr   The string from which to extract that item
  *
  * @return       The extracted string
  */
  private String NthItemOf(char delim, int whom, String aStr)  {
    char xCh;
    int nCh = (int)(delim), lxx = StrLength(aStr), here, aCh;
    boolean xact = true, wordy = false, seen = false;
    String aWord = "";
    if (nCh >= 10) if (nCh <= 32)  {
      if (nCh == 32) wordy = true;
      xact = false;} //~if
    whom--; // comes in 1-based, now 0-base
    try {  if (whom >= 0) for (here=0; here <= lxx-1; here++) {
      xCh = CharAt(here,aStr);
      aCh = (int)(xCh);
      if (aCh <= 0) break;
      if (xact)  {
        if (aCh == nCh)  {
          whom--;
          if (whom<0) break;
          continue;}} //~if
      else if (aCh <= 32)  {
        if (wordy)  { // one or more spaces..
          if (seen) whom--;
          seen = false;
          if (whom<0) break;
          continue;} //~if
        else if (aCh < 32)  { // any ctl counts as \n or \r
          whom--;
          if (whom<0) break;
          continue;}} //~if
      else seen = true;
      if (whom <= 0) aWord = aWord + xCh;} //~for
    } catch (Exception ex)  {}
    return aWord;} //~NthItemOf

 /**
  * Converts an integer to String (for JS, Java not always auto'ly coerces).
  *
  * @param  whom  The integer
  *
  * @return       The same number as String
  */
  private String CvInt2Str(int whom)  { // 'cuz JS is untyped
    return "" + whom;} //~CvInt2Str

  // Widget public methods..

/**
  * Pack up to 4 String bytes into an integer. Bytes are packed
  * "Big Endian" or "Little Endian" according to the parameter BigE.
  * Short strings may be space-filled or zero-filled; space-filled
  * also converts non-ASCII to spaces.
  *
  * @param  whom    The string
  * @param  BigE    True: Big Endian packing order
  * @param  spcf    True: space fill
  *
  * @return         The number
  */
  public static int Pack42i(String whom, boolean BigE, boolean spcf)  {
    int nx, info, lxx = StrLength(whom), rez = 0;
    for (nx=0; nx <= 3; nx++) {
      if (BigE) info = nx;
        else info = 3-nx;
      if (info<lxx) info = (int)(CharAt(info,whom))&127;
      if (((info-32)&255)>94)  {
        if (spcf) info = 32;
        else if (info<8) info = 0;} //~if
      rez = (rez<<8)+info;} //~for
    return rez;} //~Pack42i

 /**
  * Convert up to 4 bytes packed in an integer to String. Bytes are packed
  * "Big Endian" or "Little Endian" according to the parameter BigE.
  * Non-ASCII bytes are omitted (ignored).
  *
  * @param  whom    The number
  * @param  BigE    True: Big Endian packing order
  *
  * @return         The string
  */
  public static String UnPak4i(int whom, boolean BigE)  {
    String aLine = "";
    int nx, info;
    for (nx=0; nx <= 3; nx++) {
      info = whom;
      if (BigE)  {
        info = info>>24;
        whom = whom<<8;} //~if
      else whom = whom>>8;
      info = info&255;
      if (info>31) if (info<127)
        aLine = aLine + ((char)info);
      } //~for
    return aLine;} //~UnPak4i

 /**
  * Returns the reference number for this widget = its index in AllWgts[].
  *
  * @return       The reference number = RefNum
  */
  public int GetRefNum() {return RefNum;} //~GetRefNum

 /**
  * Returns the type of this widget with its flags.
  *
  * @return       The type = Gtype
  */
  public int GetType() {return Gtype&0xFFFF;} //~GetType

 /**
  * Returns the animation bits stored in the upper half of the type.
  *
  * @return       The animation bits = Gtype>>16
  */
  public int GetAnimBits() {return (Gtype>>16)&0xFFFF;} //~GetAnimBits

 /**
  * Returns the information value for this widget.
  *
  * @return       The information value = Ginfo
  */
  public int GetInfo() {return Ginfo;} //~GetInfo

 /**
  * Returns the specified color for this widget, = 0x00RRGGBB or -1 if none.
  * From a sprite this returns the color of the top (last) contained rect,
  * icon, or text (if any).
  *
  * @return       The specified color = Gcolor
  */
  public int GetColor() {return Gcolor;} //~GetColor

 /**
  * Returns the reference number of this widget's parent.
  *
  * @return       The Parent
  */
  public int GetParent() {return Parent;} //~GetParent

 /**
  * Returns the position of this widget in its parent, packed into one int.
  *
  * @return       The position = (Gtop<<16)+Gleft
  */
  public int GetPosn() {return (Gtop<<16)|Gleft&0xFFFF;} //~GetPosn

 /**
  * Returns the size of this widget's containing rectangle.
  *
  * @return       The size = (Gtall<<16)+Gwide
  */
  public int GetSize() {return (Gtall<<16)+Gwide;} //~GetSize

 /**
  * Returns the text line from a GameTxLn widget, otherwise "".
  *
  * @return       The text line = Gdata as a String
  */
  public String GetTextLn() {return "";} //~GetTextLn // only works in GameTx

 /**
  * Returns the pixel width of a GameTxLn widget's text, otherwise 0.
  *
  * @return       The width of the whole text line
  */
  public int GetTextWide() {return 0;} //~GetTextWide // only works in GameTx

 /**
  * Returns the selection of a GameTxLn or scroll limits of a GameScro,
  * otherwise 0. This method is defined here in the base class so you
  * don't need to bother with casting the GameWgt as a GameTxLn or GameScro
  * when the override mechanism automatically does that for you.
  *
  * @return       The selection or scroll limits = Gselim
  */
  public int GetSelim() {return Gselim;} //~GetSelim // only in GameTx,GameScr

 /**
  * Safely (no exceptions) returns an indexed value from this widget's Gdata,
  * or else a specified default value if it doesn't exist.
  *
  * Use an index here=-1 and a default value also defalt=-1 to safely get
  * the array size, which will be >0 if the array exists, otherwise =0
  *
  * @param  here    The index
  * @param  defalt  The value to return if Gdata[here] does not exist
  *
  * @return         The indexed value = Gdata[here]
  */
  public int GetIxData(int here, int defalt) {
    if (here<0) if (here==defalt) return SafeArySz(Gdata);
    return SafeAryElt(here,Gdata,defalt);} //~GetIxData

 /**
  * Safely (no exceptions) sets an existing indexed value in this widget's
  * Gdata. This does not enlarge the array if it doesn't exist.
  *
  * @param  here    The index
  * @param  whom    The value to set it to
  */
  public void SetIxData(int here, int whom) {
    if (Gdata==null) return;
    if (here >= 0) if (here < Gdata.length)
      Gdata[here] = whom;} //~SetIxData

 /**
  * Set the color for this widget, = 0x00RRGGBB or -1 if none.
  * In a sprite this sets the color of the top (last) contained rect,
  * icon, or text (if any).
  *
  * @param  color    The color to set Gcolor to, <0 is no color
  */
  public void SetColor(int color) {Gcolor = color;} //~SetColor

 /**
  * Set the selection of a GameTxLn or scroll limits of a GameScro.
  * This method is defined here in the base class so you don't need
  * to bother with casting the GameWgt as a GameTxLn or GameScro
  * when the override mechanism automatically does that for you.
  *
  * @param  whar   The new selection or scroll limits for Gselim
  */
  public void SetSelim(int whar) {Gselim = whar;} //~SetSelim // GameTx,GameScr

 /**
  * Set the cursor in a GameTxLn's text, either as a pixel offset, or else
  * as calculated from a character offset.
  *
  * @param  pix    True if the cursor position is a pixel offset
  * @param  whar   The new cursor position as a pixel or character offset
  */
  public void SetBlinker(boolean pix, int whar) {} // only works in GameTx

 /**
  * Set a GameTxLn's text. This method is defined here in the base class
  * so you don't need to bother with casting the GameWgt as a GameTxLn
  * when the override mechanism automatically does that for you.
  *
  * @param  aLine    The text to set Gdata to
  */
  public int PutTextLn(String aLine) {return 0;} //~PutTextLn // only in GameTx

 /**
  * Set the type of this widget, but only if previously undefined = -1.
  * This can only be called once, when the widget is created.
  *
  * @param  info    The type number to set Gtype to
  */
  public void SetWgType(int info) {
    if (Gtype+1 ==0) Gtype = info;} //~SetWgType

 /**
  * Set the Parent for this widget.
  * It should be a valid RefNum of some other widget, or else zero.
  *
  * @param  whom    The RefNum to set Parent to, 0 is none
  */
  public void SetParent(int whom) {
    if (whom != RefNum) Parent = whom;} //~SetParent

 /**
  * Set the flag bits in this widget's Gtype. The type itself is unchanged.
  *
  * @param  info    The flag bits to set Gtype to
  */
  public void SetFlags(int info) {
    Gtype = Gtype&~0xFF00|info&0xFF00;} //~SetFlags

 /**
  * Set the information value of this widget, and clear the animation bits
  * in Gtype. The animated widgets GameIcon and GameSprit use the upper
  * half of Gtype, but changing Ginfo could invalidate current animation
  * parameters, so clearing those bits forces a restart at its beginning.
  *
  * @param  info    The information value to set Ginfo to
  */
  public void SetInfo(int info) {
    if (Gtype>0xFFFF) Gtype = Gtype&0xFFFF; // restarts any animation
    Ginfo = info;} //~SetInfo

 /**
  * Test the information value of this widget and return true if it is
  * a credible 3- or 4-character ID code, false if not. 3-character ID
  * codes are not used in JavaGame, but if they were, they must begin
  * with a letter. 4-character ID codes must be all ASCII characters
  * in the range '!'..'~'. JavaGame mostly uses only letters.
  *
  * @return     True if this has a credible 3- or 4-character ID code
  */
  public boolean IsValidIDtxt() { // T: macro-gen'd click target is "xxxx"
    int whom = Ginfo, info = -1;
    if (whom<0x410000) return false; // not a valid ID if less than 3ch..
    if ((whom&0x80808080) !=0) return false;
    if (((whom+0x01010101)&0x80808080) !=0) return false;
    if (whom<0x21000000) info = 0xFFFFFF;
    if ((((whom&info&0x60606060)+(info&0xE0E0E0E0))&0x80808080) !=0)
      return false;
    return true;} //~IsValidIDtxt

 /**
  * Return the information value of this widget as a String, or else ""
  * if not valid.
  *
  * @return      The 3- or 4-character ID code
  */
  public String GetIDtxt() { // retn macro-generated click target as "xxxx"
    int whom = Ginfo, info = -1, lxx = 0, why = 0;
    String rez = "";
    GameWgt prnt = null;
    while (true) {
      why++; // why = +1 (+2*lxx)
      if (lxx==0) if (!IsValidIDtxt()) break; // not a valid ID..
      why++; // why = +2
      if (whom>0) rez = UnPak4i(whom,true);
        else break; // (can't, already checked)
      why = 0;
      break;} //~while
    if (Parent>0) if (why==1) { // Ginfo is otherwise occupied, try 1 up..
      prnt = GetIxWgt(Parent);
      if (prnt != null) if (prnt.GetInfo()>0x410000) // if (false)
        if ((prnt.GetType()&255)==5) rez = prnt.GetIDtxt();} //~if
    return rez;} //~GetIDtxt

 /**
  * Return the class name and RefNum for this widget, plus its ID code if any.
  *
  * @return     The reference String for this widget
  */
  public String toString() { // . . . . . . . . . . . . . . . . . // (GameWgt)
    String msg = GetIDtxt();
    if (StrLength(msg)==0) msg = ")";
      else msg = ") '" + msg + "'";
    return "(GameWgt #" + RefNum + msg;} //~toString

 /**
  * Returns the vertical position of this widget = its top pixel row.
  *
  * @return       The vertical position = Gtop
  */
  public int GetVposn() {return Gtop;} //~GetVposn

 /**
  * Returns the horizontal position of this widget = its leftmost pixel column.
  *
  * @return       The horizontal position = Gleft
  */
  public int GetHposn() {return Gleft;} //~GetHposn

 /**
  * Returns the height of this widget.
  *
  * @return       The height = Gtall
  */
  public int GetTall() {return Gtall;} //~GetTall

 /**
  * Returns the width of this widget.
  *
  * @return       The width = Gwide
  */
  public int GetWide() {return Gwide;} //~GetWide

 /**
  * If this is a sprite, returns its vertical velocity (positive is down)
  * as a (double) number of pixels per frame.
  *
  * @return       The high 16 bits of the information value Ginfo
  */
  public double GetDnVelo() {
  if ((Gtype&255) !=6) return 0.0;
  return (Ginfo>>16)/256.0;} //~GetDnVelo

 /**
  * If this is a sprite, returns its horizontal velocity (negative is left)
  * as a (double) number of pixels per frame.
  *
  * @return       The low 16 bits of the information value Ginfo
  */
  public double GetRtVelo() {
  if ((Gtype&255) !=6) return 0.0;
  return SignExtend(Ginfo)/256.0;} //~GetRtVelo

 /**
  * Sets the vertical position of this widget = its top pixel row.
  *
  * @param  vert   the new vertical position
  */
  public void SetVposn(int vert) {Gtop = vert;} //~SetVposn

 /**
  * Sets the horizontal position of this widget = its leftmost pixel column.
  *
  * @param  horz   the new horizontal position
  */
  public void SetHposn(int horz) {Gleft = horz;} //~SetHposn

 /**
  * Sets the height of this widget.
  *
  * @param  tall   the new height, if >0
  */
  public void SetTall(int tall) {if (tall>0) Gtall = tall;} //~SetTall

 /**
  * Sets the width of this widget.
  *
  * @param  wide   the new width, if >0
  */
  public void SetWide(int wide) {if (wide>0) Gwide = wide;} //~SetWide

 /**
  * If this is a sprite, sets its vertical velocity (positive is down)
  * as a (double) number of pixels per frame. The internal representation
  * is fixed-point, so any velocity smaller than 0.0025 pixels per frame
  * is effectively zero, and larger than 127.99 is clipped to that value.
  *
  * @param  velo   the new vertical velocity
  */
  public void SetDnVelo(double velo) {
    int info = 0;
    if ((Gtype&255) !=6) return;
    if (velo>128.0) velo = 128.0;
    else if (velo+128.0<0.0) velo = -128.0;
    info = (int)Math.round(velo*256.0);
    if (info>0x7FFF) info = 0x7FFF;
    Ginfo = (info<<16)|Ginfo&0xFFFF;
    if (Logging) if (GoodOptn) System.out.println("(SetDnVel #" + RefNum + ") "
        + velo + " => " + info + SeeHex(" = ",Ginfo,""));} //~SetDnVelo

 /**
  * If this is a sprite, sets its horizontal velocity (negative is left)
  * as a (double) number of pixels per frame. The internal representation
  * is fixed-point, so any velocity smaller than 0.0025 pixels per frame
  * is effectively zero, and larger than 127.99 is clipped to that value.
  *
  * @param  velo   the new horizontal velocity
  */
  public void SetRtVelo(double velo) {
    int info = 0;
    if ((Gtype&255) !=6) return;
    if (velo>128.0) velo = 128.0;
    else if (velo+128.0<0.0) velo = -128.0;
    info = (int)Math.round(velo*256.0);
    if (info>0x7FFF) info = 0x7FFF;
    Ginfo = (Ginfo&-0x10000)|info&0xFFFF;
    if (Logging) if (GoodOptn) System.out.println("(SetRtVel #" + RefNum + ") "
        + velo + " => " + info + SeeHex(" = ",Ginfo,""));} //~SetRtVelo

 /**
  * Set the dimensions of this widget's containing rectangle.
  * The specified dimensions must be >0, but if you specify tall=0,
  * it will preserve the previous height and change only the width.
  * If you specify wide=0 and tall is a packed size (such as returned
  * from GetSize) then it will use that packed size. The size of the
  * game window (RefNum=0) are also captured locally for future use.
  *
  * @param  tall   the new height
  * @param  wide   the new width
  */
  public void SetSize(int tall, int wide) {
    if (tall==0) tall = Gtall; // OK to change only wide
    if (wide==0) if (tall>0xFFFF) { // OK to pass packed in tall..
      wide = SignExtend(tall);
      tall = tall>>16;} //~if
    if (tall <= 0) return;
    if (wide <= 0) return;
    if (Logging) if (MaxLogWgts>0) if (MaxLogWgts>Max_Log-8) {
      System.out.println("(SetSz #" + RefNum + ") " + tall + "/" + wide);
      MaxLogWgts--;} //~if
    Gtall = tall;
    Gwide = wide;
    if (RefNum>0) return; // get game dims from 1st wgt..
    if (wide<8) return;
    if (wide>4095) return;
    if (tall<8) return;
    if (tall>4095) return;
    WinHi = tall;
    WinWi = wide;
    MaxWinSize = (tall<<16)+wide;} //~SetSize

 /**
  * Get this widget's Gdata array.
  *
  * @return        a reference to the array in Gdata
  */
  public int[] GetData() {
    return Gdata;} //~GetData

 /**
  * Replace this widget's Gdata array.
  *
  * @param  data   the new array for Gdata
  */
  public void SetData(int[] data) {
    int tye = Gtype&255;
    if (tye>0) if ((tye !=3)||(data != null))
      Gdata = data;} //~SetData

 /**
  * This is an internal utility for formatting a fixed-point animation value.
  *
  * @param  before  a prefix to add to the front
  * @param  whom    the value with an 8-bit fractional part
  *
  * @return         the composite display value
  */
  public String FixHaff(String before, int whom) { // from LogMe->FmtInfo
    return FormFixt(before,whom);} //~FixHaff

 /**
  * This is an internal utility for formatting large numbers as hex.
  *
  * @param  before  a prefix to add to the front
  * @param  whom    the value with an 8-bit fractional part
  * @param  after   a suffix to add to the end
  *
  * @return         the resulting display value
  */
  public String HexyMore(String before, int whom, String after) {
    return HexIfMore(before,whom,after);}

 /**
  * This is an internal utility for formatting a widget's information value.
  * Some subclasses override this to get different formatting.
  *
  * @param  before  a prefix to add to the front
  * @param  tye     the widget's type
  *
  * @return         the composite display value
  */
  public String FmtInfo(String before, int tye) { // o'ride for packed data
    if ((tye==0)||(tye==5)) if (IsValidIDtxt())
      return before + "`" + GetIDtxt() + "`";
    if (tye>0) if (tye<6) return before + (Ginfo&0x3FFF);
    return before + Ginfo;} //~FmtInfo // caller must add space after

 /**
  * This is an internal utility for formatting the front part of a widget's
  * display line in LogMe.
  *
  * @param  before  a prefix to add to the front
  * @param  dimz    true to show also the widget's size
  * @param  after   a suffix to add to the end
  *
  * @return         the composite display value
  */
  public String FmtFront(String before, boolean dimz, String after) {
    String msg = "Wgt #"; // only called locally; after needs prefix spc
    int posn = GetPosn(), whar = AddPrntsPos(posn), t1 = (Gtype>>16),
        t2 = (t1>>4)&15, tye = Gtype&255;
    msg = NthItemOf(' ',(Gtype&7)+1, // we only do the standard items for now..
        "wRect wPixM wIcon wFont wTxLn wList wSpri wScro") + " #";
    t1 = t1&15;
    msg = msg + RefNum + ") t=" + tye + IffyStr((Gtype&0xFF00)==0,"",
        "+" + ((Gtype>>8)&255)) + IffyStr((Gtype&0xFF0000)==0,"","-t1="
        + ((Gtype>>16)&15)) + IffyStr((Gtype&0xFF000000)==0,"","-t2="
        + ((Gtype>>20)&15));
    if (dimz) {
      msg = msg + " " + Gtop + "/" + Gleft;
      if (whar != posn)
        msg = msg + " (" + (whar>>16) + "/" + SignExtend(whar) + ")";
      msg = msg + " +" + Gtall + "/" + Gwide + " p=" + Parent;} //~if
    // => B4 'wNm +rn) t=ty+fg-an {tp/lf (abs) +ta/wi p=pr} c=co i=in' aft
    msg = before + msg + FmtInfo(SeeHex(" c=",Gcolor," i="),tye)
        + IffyStr(Gselim==0,"",IffyStr(Gselim>255,SeeHex(" s=x",Gselim,""),
          " s=" + Gselim)) + IffyStr(dimz,""," (~" + MaxLogWgts + ")");
    return msg + after;} //~FmtFront

 /**
  * This prints a widget's display line in the console log.
  *
  * @param  before   a prefix to add to the front of the line
  */
  public void LogMe(String before) { // caller ensures if (Logging)
    String msg = before + FmtFront("(",true," ["), aStr;
    boolean more = false, linx = ((Gtype-5)&255)<3;
    int nx, zx, lxx = SafeArySz(Gdata), sofar = StrLength(msg);
    if (lxx>0) {
      if (linx) {
        if (lxx>123) {
          more = true;
          lxx = 33;}} //~if
      else if (lxx>6) {
        more = true;
        lxx = 6;} }
    if (lxx>0) {
      for (nx=0; nx <= lxx; nx++) {
        aStr = HexyMore(" ",SafeAryElt(nx,Gdata,0),"");
        zx = StrLength(aStr);
        if (sofar+zx>80) {
          if (!linx) break;
          msg = msg + "\n   +" + nx + ":";
          sofar = 7;} //~if
        msg = msg + aStr;
        sofar = sofar+zx;} //~for
      msg = msg + IffyStr(more,"..]"," ]");} //~if
    else msg = msg + "]";
    System.out.println(msg);} //~LogMe

 /**
  * CharSetup takes a position, the ASCII (numeric code of a) character,
  * a color and Parent RefNum to alter this widget to resemble an icon
  * positioned within the parent container, so when it is drawn, the
  * character appears there. This is repeated for each character in a
  * GameTxLn's line of text.
  *
  * This only works in GameFont, but is a stub so that downcast is unneeded.
  *
  * @param  top    the vertical position for this character
  * @param  left   the horizontal position for this character
  * @param  xCh    the character to be drawn there
  * @param  color  its color, 0x00RRGGBB
  * @param  whom   the parent (group) widget to insert this icon under
  * @param  logy   true to spit out many lines of console text
  *
  * @return        the pixel position ended = left+Gwide, or else err<0
  */
  public int CharSetup(int top, int left, int xCh, int color, int whom, boolean logy) {
    int rez = 0; return rez;} //~CharSetup

  /**
  * Hide or show this widget by moving it off- or on-screen. This is done
  * by adding or removing the 0x4000 bit in its vertical position, so if the
  * widget is not visible for other reasons, calling HideSho(true) probably
  * won't make it visible.
  *
  * @param  sho   true to show the widget, false to hide it
  */
  public void HideSho(boolean sho) {
    if (Gtop<0) return;
    if (sho) Gtop = Gtop&0x3FFF;
      else Gtop = Gtop|0x4000;} //~HideSho

  /**
  * This draws the default widget on the window. Some subclasses override
  * this to alter the appearance. The boundaries of a masking rectangle
  * limit the drawing to (usually) the containing widget's dimensions.
  *
  * @param  top    do not draw above this position
  * @param  left   do not draw to the left of this position
  * @param  tall   do not draw below this position +top
  * @param  wide   do not draw to the right of this position +left
  */
  public void  DrawMe(int top, int left, int tall, int wide) {    // (GameWgt)
    // constrain to given (global) edges  // o'ride for all other Gtypes // ***
    int botm = top, rite = left,
        nx = AddPrntsPos( GetPosn()),  frax = Gtype>>16, blnk = Ginfo>>14, here, thar = Ginfo&0x3FFF,
        myTop = nx>>16, myLft = SignExtend(nx), colo = Gcolor,
        self = RefNum<<3, rx = myTop+Gtall, cx = myLft+Gwide, tye = Gtype&255,
      kx, zx = 0, lxx = 0, seen = 0, logy = 0, info = 0, why = 0;
    GameWgt nolo = null,  whom,
        funt = null;  // (GameFont)
    String msg = "";                            // flgs = (Gtype>>8)&255,
    boolean more = false;
    int[] data = Gdata;
    // CallersID = 0;
    while (true) { // tye: "wRect wPixM wIcon wFont wTxLn wList wSpri wScro"
      why++; // why = 1
      why = ~why; // why = -2
      if (Gtop>0x3FFF) break; // hidden
      why--; // why = -3
      if (Ginfo+1 ==0) break; // Ginfo=-1: informational data not drawn
      why--; // why = -4
      if (top<0) break;
      why--; // why = -5
      if (left<0) break;
      why = -why; // why = 5
      if (tall <= 0) break;
      why++; // why = 6
      if (wide <= 0) break;
      if (logy==1234) // do this now, JS needs zx..
        SeenLog = SeenLog|zx;
      if (myTop>top) top = myTop;
      if (myLft>left) left = myLft;
      if (rx <= botm+tall) botm = rx-1; // was: botm = orig top
        else botm = botm+tall-1;
      if (cx <= rite+wide) rite = cx-1; // was: rite = orig left, now: edge
        else rite = rite+wide-1;
      if (logy>999)
        msg = msg + " => " + top + "/" + left + " [+" + tall
            + "/" + wide + "] (" + botm + "/" + rite + ")";
      why = ~why; // why = -7
      if (left>rite) more = true;
        else why--; // why = -8
      if (top>botm) more = true;
      if (more) break; // why = -7/-8
      lxx = SafeArySz(data);
      if (colo >= 0) if (tye !=1) { // got a color to paint it..
        more = true;
        zx = (blnk>>2)&15; // diagonal line..
        kx = 0;
        if (zx>13) { // triangular fill..
          if (Gtall<Gwide) kx = Gtall;
            else kx = Gwide;
          zx = -zx;} //~if
        if (zx<Gtall) if (zx<Gwide) { // (if too wide, just fill it)
          if (logy==1111)
            msg = msg + SeeHex(" i=",Ginfo," ") + zx;
          if ((Ginfo&0xFFE0FFC0) == 0x44604440) // 'DgD*' // thar=Ginfo&0x3FFF..
              if ((tye==0)&&((thar&255)==0x6C)||(tye==5)&&((thar&28) == 20)) {
            why = 0;                  // 'D*Dl'             // 'D*DT'/'D*DW'
            more = (Ginfo&0x00100000)==0; // T: Dd/Dg/Dn (down) not Du/Dp (up)
            if (logy>0) msg = msg + IffyStr(more," D\\"," D/")
                + zx + IffyStr(Gtall<Gwide,"-","|");
            if (zx !=0) { // 0-width line counts as "done"
              if (kx>0) {
                lxx = kx;
                if ((zx&1)==0) kx = -kx; // 'n'=0: under
                zx = 0;} //~if
              else lxx = zx-1; // = line width
              if (Gtall<Gwide) { // landscape...........................
                tall = myTop+myTop+Gtall-1;
                // wide = Gwide-1;
                nx = Gtall;
                while (Gwide>nx) { // calculate bevel..
                  if (zx==0) break;
                  nx = nx<<1;
                  zx = zx>>1;} //~while
                myTop = myTop-zx;
                if (zx==0) seen = Gtall-1;
                  else seen = Gtall+zx+zx;
                seen = ((seen<<8)+(Gwide>>1))/Gwide; // = step
                if (logy>0) if (GoodDiag) msg = msg + " (" + myTop + "/" + myLft
                    + " +" + seen + ", " + tall + "/" + (tall-myTop) + ")";
                myTop = myTop<<8;
                for (cx=cx-1; cx >= myLft; cx += -1) { // each column.. (cx = myLft+Gwide)
                  rx = myTop>>8;
                  if (more) rx = tall-rx; // dn: step frm botm to top ea coln
                  for (nx=lxx; nx >= 0; nx += -1) {
                    if (logy>0) if (GoodDiag)
                      msg = msg + " " + rx + "/" + cx;
                    if (rx >= top) if (rx<botm) if (cx >= left) if (cx<rite)
                      PokePixel(colo,rx,cx);
                    if (kx<0) rx++;
                    else if (kx>0) rx--;
                    else if (more) rx--;
                      else rx++;} //~for
                  myTop = myTop+seen;}} //~if //~for (each column) (landscape)
              else { // portrait.........................................
                wide = myLft+myLft+Gwide-1;
                // tall = Gtall-1;
                nx = Gwide;
                while (Gtall >= nx) { // calculate bevel..
                  if (zx==0) break;
                    nx = nx<<1;
                    zx = zx>>1;} //~while
                myLft = myLft-zx;
                if (zx==0) seen = Gwide-1;
                  else seen = Gwide+zx+zx;
                seen = ((seen<<8)+(Gtall>>1))/Gtall; // = step
                if (logy>0) if (GoodDiag) msg = msg + " (" + myTop + "/" + myLft
                    + " +" + seen + ", " + wide + "/" + (wide-myLft) + ")";
                myLft = myLft<<8;
                for (rx=rx-1; rx >= myTop; rx += -1) { // each row, bottom to top..
                  cx = myLft>>8;
                  if (more) cx = wide-cx; // dn: step frm right to left ea row
                  for (nx=lxx; nx >= 0; nx += -1) {
                    if (logy>0) if (GoodDiag)
                      msg = msg + " " + rx + "/" + cx;
                    if (rx >= top) if (rx<botm) if (cx >= left) if (cx<rite)
                      PokePixel(colo,rx,cx);
                    if (more) {
                      if (kx>0) cx++;
                        else cx--;} //~if
                    else if (kx>0) cx--;
                      else cx++;} //~for
                  myLft = myLft+seen;}}} //~if //~for (each row) (portrait)
            more = false;}} //~if // (diagonal line) (if too wide)     ..(0-width)
        zx = 0;
        if (tye !=6) zx++;
        else if (lxx <= 0) zx++;
        if (more) if (zx>0) {
          RectFill(colo,top,left,botm,rite);}} //~if
      why--; // why = -9
      rx = RefNum;
      if ((tye&-4)==4) { // got a list of components to draw..
        tall = botm+1-top;
        wide = rite+1-left;
          if (logy>999) msg = msg + " * " + tall + "/" + wide;
          if (logy>0) msg = msg + " *" + lxx;
        if (lxx>0) for (nx=0; nx <= lxx-1; nx++) {
          info = SafeAryElt(nx,data,rx);
          if (info==rx) continue; // shouldn't hap
          if (info>0) whom = GetIxWgt(info);
            else whom = nolo;
          if (logy>0)
            if ((whom != nolo)||(seen<3))
              msg = msg + " #" + info + IffyStr(whom==nolo,"f","T");
          if (whom==nolo) {
            seen++;
            continue;} //~if
          info = whom.  Gtype&255;
          if (info != 1) if (info != 3) {
             zx = WhoCalled(rx);
            whom.  DrawMe(top,left,tall,wide);}} //~for
        break;} //~if // why = -9
      why--; // why = -10
      if (rx !=0) break; // nothing else to do..
      why = -why; // why = 10
      if (tye !=1) break;
      why++; // why = 11
      if (lxx <= 0) break;
      why++; // why = 12
      if ((myTop|myLft) !=0) break;
      why++; // why = 13 // otherwise draw main game board..
      if (Ginfo <= 0) break;
      whom = GetIxWgt(Ginfo);
      why++; // why = 14
      if (whom==nolo) break;
      zx = whom.Gtype&255;
      why++; // why = 15
      if (zx !=5) break;
       zx = WhoCalled(RefNum); // =0
      whom.  DrawMe(top,left,tall,wide);
      if (Logging) {
        zx =  JavaGame.  DrawLogger;
        if (zx>0) if (zx<8) if ((zx&1) !=0) {
          if (logy==0) logy++; // log that we are drawing after user event..
          why++; // why = 16
          break;}} //~if
      why = 0;
      break;} //~while
    } //~DrawMe

/**
  * Give this widget focus, or else take focus away.
  * A widget that is not typeable, clickable, or dragable cannot accept focus.
  * A GameScro widget never accepts focus.
  *
  * @param  nuly   true to give focus, false to take it
  * @param  vert   the vertical position of a mouse-click giving focus
  * @param  horz   the horizontal position of the mouse-click
  *
  * @return        true if focus accepted, false if can't or losing it
  */
  public boolean SetFocus(boolean nuly, int vert, int horz) {
    int tye = Gtype&255;
    if (!nuly) return false;
    if (tye==7) return false;
    if (Ginfo+1 ==0) return false;
    if ((Gtype&0xE00)==0) return false;
    if ((Gtype&0x2000) !=0) return false;
    // if (tye==5) if (SafeArySz(Gdata)>0) // FocEv sets this wgt marching ants
    return true;} //~SetFocus

 /**
  * Convert a packed position of this widget into an absolute window offset.
  * The parameter here might come from the widget's GetPosn to get its
  * absolute position in the window.
  *
  * @param  here    The widget-relative offset (V<<16)+H in the parent widget
  *
  * @return         The packed window-relative offset (V<<16)+H
  */
  public int AddPrntsPos(int here) { // recursive to top
    int zx, fust, deep = WhoCalled(0);
    GameWgt whom;
    if (deep>0) {
      // CallersID = 0;
      deep = 0;} //~if
    fust = (-deep)&15;
    zx = WhoCalled(-fust);
    if (Parent>0) if (Parent != RefNum) if (fust<8) { // never go that deep
      zx = WhoCalled(-fust-1); // CallersID--;
      whom = GetIxWgt(Parent);
      if (whom != null)
        here = whom.AddPrntsPos(AddPair(whom.GetPosn(),here));
      zx = WhoCalled(0); // CallersID++;
      zx = WhoCalled(zx+1);} //~if
    if (fust==0) {
      if (false) if (deep+999<0) System.out.println(LogVH("  => ",here,""));
      zx = WhoCalled(0);} //~if // CallersID = 0;
    return here;} //~AddPrntsPos

 /**
  * Return true if this widget's rectangle is at least partly on-screen.
  * If the widget is not visible for other reasons, IsVisible might return
  * a false positive.
  *
  * @return        true the widget's rectangle is at least partly on-screen
  */
  public boolean IsVisible() { // T: rect is on-screen (may not be visi..
    int posn = AddPrntsPos(GetPosn()), // ..if transparent or under other wgt)
        topper = GetSize()+posn-0x20002, butter = MaxWinSize-0x20002-posn,
        logy = 0;
    if (Logging) if ((logy&0xFFFF)==RefNum) System.out.println("(IsVisi #" + RefNum
        + LogVH(") t=",topper,LogVH(" b=",butter,LogVH(" s=",MaxWinSize,""))));
    if ((topper&0x80008000) !=0) return false;
    if ((butter&0x80008000) !=0) return false;
    return true;} //~IsVisible

 /**
  * Search for a value in this widget's Gdata array. Returns the index
  * position of the first such item found, or -1 if not there.
  *
  * @param  valu   the value to look for in Gdata
  *
  * @return        the index position of the first such item found, or -1
  */
  public int FindElemt(int valu) { // rtns index, or -1 if not there
    int nx, lxx = SafeArySz(Gdata)-1, info = valu+1;
    if (((Gtype-5)&255)<3) for (nx=0; nx <= lxx; nx++) // else not a valid group type
      if (SafeAryElt(nx,Gdata,info)==valu) return nx;
    return -1;} //~FindElemt

 /**
  * Delete a value in this widget's Gdata array given its index.
  * The remaining items are shifted down to fill the gap, with zero-fill
  * at the end. It is assumed that all the non-zero data is at the front,
  * so a run of zeroes terminates the shifting down.
  *
  * @param  inx   the index position of the item to delete
  */
  public void DeleElemt(int inx) { // ***
    int nx, info, prio = SafeAryElt(inx,Gdata,0),
        seen = 0, lxx = 0, why = 0;
    int[] data = Gdata;
    while (true) { // once thru..
      why++; // why = 1
      if (data == null) break;
      why++; // why = 2
      if (((Gtype-5)&255)>2) break; // not a valid group type
      lxx = SafeArySz(data)-1;
      why++; // why = 3
      if (inx<0) break;
      why++; // why = 4
      if (inx>lxx) break; // index off end
      for (nx=inx; nx <= lxx; nx++) {
        info = 0;
        if (nx < data.length-1)
          info = data[nx+1];
        data[nx] = info;
        if (info==0) {
          if (seen>3) break; // 4 zeroes in a row: probly end
          seen++;
          continue;} //~if
        seen = 0;} //~for
      why = 0;
      break;} //~while
    if (Logging)
      System.out.println("(DelElt #" + RefNum + ") #" + prio + " @"
          + inx + " = " + why);} //~DeleElemt

 /**
  * Exchange two adjacent values in this widget's Gdata array given the
  * index. The remaining items are not affected. Zeroes can only be
  * moved up, not down.
  *
  * @param  inx   the index position of the first of the two items
  */
  public void XchElemt(int inx) {
    int prio = SafeAryElt(inx+1,Gdata,0), info = 0, why = 0;
    int[] data = Gdata;
    while (true) { // once thru..
      why++; // why = 1
      if (data == null) break;
      why++; // why = 2
      if (((Gtype-5)&255)>2) break; // not a valid group type
      // lxx = SafeArySz(data);
      why++; // why = 3
      if (inx<0) break;
      why++; // why = 4
      // if (inx+2>lxx) break; // index off end
      if (inx > data.length-2)
        break;
      why++; // why = 5
      if (prio==0) break;
      info = data[inx];
      data[inx+1] = info;
      data[inx] = prio;
      why = 0;
      break;} //~while
    if (Logging) System.out.println("(XchElt #" + RefNum + ") " + info
        + " / " + prio + " @" + inx + " = " + why);} //~XchElemt

 /**
  * Insert a new value at the end of (the non-zero values in) this
  * widget's Gdata. Returns the index position where it was inserted,
  * or -1 if it failed. If Gdata is too small, it will be enlarged, but
  * if it is null, AddElemt will fail. If it is already there, AddElemt
  * simply returns its index (same as FindElemt).
  *
  * @param  valu   the value to look for in Gdata
  *
  * @return        the index position of the first such item found, or -1
  */
  public int AddElemt(int valu) { // Gdata must exist, retns FindElt
    int nx, info, more = 0, lxx = 0, why = 0;
    int[] nuly = null;
    int[] data = Gdata;
    while (true) { // once thru..
      why++; // why = 1
      if (data == null) break;
      why++; // why = 2
      if (((Gtype-5)&255)>2) break; // not a valid group type
      lxx = SafeArySz(data)-1;
      why++; // why = 3
      for (nx=0; nx <= lxx; nx++) {
        info = data[nx];
        if (info !=0) if (info != valu) continue;
        data[nx] = valu;
        why = ~nx;
        break;} //~for
      if (why<0) break; // success
      why++; // why = 4
      lxx++;
      if (lxx>250) more = lxx+256;
      else if (lxx<5) more = 8;
        else more = lxx<<1;
      nuly = new int[more];
      if (nuly == null) break; // why = 4
      why++; // why = 5
      for (nx=more-1; nx >= 0; nx += -1) {
        info = 0;
        if (nx==lxx) {
          info = valu;
          why = ~nx;} //~if
        else if (nx < data.length)
          info = data[nx];
        nuly[nx] = info;} //~for
      if (why<0) SetData(nuly); // otherwise why = 5: failed
      break;} //~while
    if (why<0) {
      info = ~why;
      if (Logging)
        System.out.println("(AddElt #" + RefNum + ") @" + info
            + " +" + more + " => " + valu);
      return info;} //~if
    System.out.println("(not AddElt #" + RefNum + ") +" + more
        + " = " + why);
    return -1;} //~AddElemt

 /**
  * If this is a component of a checkbox created by the checkbox macro,
  * CheckIt (optionally) toggles the checkmark and returns the new setting.
  * This works on the containing group (which CheckIt searches its subwidgets
  * to find the actual checkmark icon) or on the checkmark icon itself, which
  * is recognized by having the same first element as the default checkmark
  * array. The checkmark is hidden by setting the pixel offset past the end
  * of the actual checkmark pixels in its array.
  *
  * @param  tog   true toggles the checkmark, false leaves it unchanged
  *
  * @return       true if the checkmark is showing, false if not
  */
  public boolean CheckIt(boolean tog) { // only works with checkbox macro // ***
    int bitz = JavaGame. ChkBitz[0], tye = Gtype&255, nx, info = 0, why = 0;
    boolean rez = false;
    GameWgt nuly = null;
    String msg = "";
    while (true) { // once thru..
      why++; // why = 1
      msg = " #" + RefNum + ") t=" + tye;
      why++; // why = 2
      if (tye==5) for (nx=0; nx <= 99; nx++) {
        info = GetIxData(nx,0);
        if (info <= 0) break;
        nuly = GetIxWgt(info);
        info = 0;
        if (nuly==null) break;
        info = nuly.GetType()&255;
        if (info==2) if (nuly.GetIxData(0,0) == bitz) {
          rez = nuly.CheckIt(tog);
          why = 4;
          break;} //~if
        info = 0;} //~for
      else if (tye==2) if (GetIxData(0,0) == bitz) {
        why++; // why = 3
        info = Ginfo;
        rez = info==0;
        if (tog) {
          if (rez) {
            rez = false;
            info = 12;} //~if
          else {
            rez = true;
            info = 0;} //~else
          Ginfo = info;} //~if
        msg = msg + IffyStr(tog," => "," =")
            + IffyStr(rez,"T i=","F i=") + info;
        break;} //~if // why = 3
      break;} //~while // (once or twice thru) // why = 2/4
    if (why==1) msg = "?)";
    if (Logging)
      System.out.println("(ChekIt" + msg + " = " + why);
    return rez;} //~CheckIt

 /**
  * Set the position of the top-left corner of this widget's containing
  * rectangle. If you specify left=0 and top is a packed position (such
  * as returned from GetPosn) but >= 0x10000, then it will use that packed
  * position. If you want to use this shortcut to set a position that might
  * be on the top row of its parent, you can pass it (newpos|0x20000000)
  * and the extra high bit will be removed.
  *
  * @param  top    the new top
  * @param  left   the new left edge
  */
  public void SetPosn(int top, int left) {
    if (left==0) if (top>0xFFFF) { // OK to pass packed in top..
      left = SignExtend(top);
      top = (top>>16)&~0x2000;} //~if
    Gtop = top;
    Gleft = left;} //~SetPosn

 /**
  * If this is a component of a radio button created by the radio button
  * macro, RadioBtn either selects a new setting or returns the current
  * setting. It is expected that this will be called for the containing
  * group, and RadioBtn searches its subwidgets to find the one to select,
  * and to deselect the others, by setting each icon widget to either show
  * or hide the selection button. The icon widgets are recognized by having
  * the same second element as the default radio button array.
  *
  * Radio buttons are distinguished by a single character (preferably a
  * digit or letter) embedded in their information ID. If you call RadioBtn
  * with a negative parameter, it will return the current setting character.
  * If you call RadioBtn with (the ASCII value of) a character greater than
  * space (and not greater than '~', that is 33..126) then it will search for
  * the button with that ID and (if found) set it to selected, and all the
  * others to deselected. If you call RadioBtn with a number 0..31, then it
  * selects whichever button has the character '0'..'9' which is that single
  * digit, or else the letter 'A'..'U' corresponding to the values 10..31
  * determined by adding 55 to the number given. If that's too confusing,
  * just use the letters.
  *
  * Internally, the icons show a hollow circle for deselected, and (later
  * in the same array) a button image for selected, the offset determines
  * which. Each button is a group over a white solid circle icon and the
  * black selection button icon. The button group ID is the 4-character
  * code "Rd_*" where the '*' is replaced by the selector character. The
  * group ID over the set of radio buttons linked together is "RdB*" where
  * again the '*' is replaced by whatever code has been selected, so that
  * asking for the current setting requires no search. The searches and
  * settings are recursive through RadioBtn.
  *
  * @param  whom  the code to be set, 33..126, or else -1 to query
  *
  * @return       the current setting code, 33..126
  */
  public int RadioBtn(int whom) { // only works w/radio btn macr     (GameWgt)
    boolean logy = Logging ;
    int bitz = JavaGame. RadBtz[1],
        tye = Gtype&255, info = Ginfo, nx, rez = 0, why = 0;
    GameWgt nuly = null;
    String msg = "";
    while (true) { // once thru (but recursive)..
      why++; // why = 1
      rez--; // rez = -1 if cannot find it
      if (logy) {
        if (whom>0) msg = SeeHex(" w=x",whom,"");
          else msg = " w=" + whom;
        msg = ") t=" + tye + FmtInfo(" i=",tye) + msg;} //~if // (FmtInf wants ty)
      why++; // why = 2
      if (tye==5) {
        why = 5;
        if ((info&0x7FFF0000) != 0x52640000) break; // = 'Rd**'
        if (((info>>8)&255) == 66) { // = 'B' (`RdB*` = group head)
          if (logy) msg = msg + " (GH) " + GetIDtxt();
          if ((whom<0)||(((whom-info)&255)==0)) { // query or btn=wanted..
            if (logy) msg = msg + " ?";
            why = -why; // why = -5 (OK res)
            rez = info;
            break;} //~if
          if (whom<10) whom = whom+0x52645F30; // => 'Rd_0'..'Rd_9'
          else if (whom<36) whom = whom+0x52645F37; // => 'Rd_A'..'Rd_Z'
          else if (whom>126) {
            why++; // why = 6 (can't set that)
            break;} //~if
          else whom = whom+0x52645F00; // => 'Rd_$'..'Rd_~'
          msg = SeeHex(" => x",whom,"");} //~if // (group head)
        for (nx=0; nx <= 99; nx++) { // each btn in grp, or each icon in btn..
          info = GetIxData(nx,0);
          if (logy) msg = msg + " " + nx + ":" + info;
          if (info <= 0) break;
          nuly = GetIxWgt(info);
          info = -info;
          if (nuly==null) break;
          info = nuly.GetType()&255;
          if (logy) msg = msg + ",t=" + info;
          if (info==5) { // this is a single btn in the grp (rtn 'on')..
            info = nuly.GetInfo();
            if (logy) msg = msg + nuly.FmtInfo(",i=",5);
            if (info==whom) {
              info = nuly.RadioBtn(whom); // turn on
              rez = info;} //~if
            else info = nuly.RadioBtn(-1); // turn it off
            if (logy) msg = msg + ",r=" + IffyStr(info<255,CvInt2Str(info),
                SeeHex("x",info,""));} //~if // why = 5
          else if (info==2) if (nuly.GetIxData(1,0) == bitz) { // = RadBtz[1]
            rez = nuly.RadioBtn(whom); // Rbtn icon: turn it on=whom or off<0
            if (logy) msg = msg + ",r=" + rez;
            why--; // why = 4
            break;} //~if
          info = 0;} //~for // (each btn in grp, or each icon in btn)
        if (logy) msg = msg + ";";
        if (rez <= 0) { // else why = 5
          why = 7;
          info = Ginfo&0x7FFFFF00;
          if (info == 0x52644200) if (rez==0) { // (`RdB*` = group head)
            why++; // why = 8
            Ginfo = info+33;
            if (logy) msg = msg + " => " + GetIDtxt();} //~if
          if (logy) msg = msg + SeeHex(" != ",whom,"");} //~if
        else if ((Ginfo&0x7FFFFF00) == 0x52644200) if (rez==whom) {
          Ginfo = whom&255|0x52644200;    // (`RdB*`) // why = 5
          if (logy) msg = msg + " => " + GetIDtxt();}} //~if // why = 5/7/8
      else if (tye==2) if (GetIxData(1,0) == bitz) { // = RadBtz[1]: btn on/off
        why++; // why = 3
        rez = 0; // info = Ginfo;
        if (whom>0) { // on: show whole btn
          rez = whom;
          info = 12;} //~if
        else info = 0; // off: show only outline
        Ginfo = info;
        if (logy) msg = msg + FmtInfo(" => i=",tye);} //~if // why = 3
      break;} //~while // (once or twice thru) // why = 2..8
    // msg = " #" + RefNum + ") t=" + tye + SeeHex(" w=",whom," i="); // (all)
    //   msg = msg + " ? = " + GetIDtxt();   // (whom<0)|(whom==info)
    //   msg = msg + " => " + GetIDtxt() + SeeHex(" != ",whom,""); // (rez<=0)
    //   msg = msg + FmtInfo(" => i=",tye);    // (btn on/off)
    if (logy) {
      if (why==1) msg = "?)"; // (can't)
      System.out.println("(RadBtn #" + RefNum + msg
          + HexIfMore(" = ", rez,"/") + why);} //~if
    return rez;} //~RadioBtn

 /**
  * This is an internal utility for generating the result that a toString
  * call to a subclass of GameWgt returns. If the parameter msg contains
  * a formatted RefNum, this replaces its prefix; otherwise it returns the
  * stated prefix with the widget's RefNum.
  *
  * @param  msg   a partial result to find the RefNum in
  * @param  whom  a prefix to replace the front of msg,
  *                 or else alone with RefNum
  *
  * @return       the composite display value
  */
  public String toStrMash(String msg, String whom) { // (internal utility)
    int nx = JavaGame. NthOfset(0," #",msg);
    if (nx>0) return whom + RestOf(nx+2,msg);  // (generates subclass toString)
    return whom + RefNum + ")";} //~toStrMash

 /**
  * Default constructor.
  */
  public GameWgt() {
    RefNum = JavaGame. NextRefNo;
    JavaGame. NextRefNo++;
    if (Logging)
      System.out.println("(new GamWgt #" + RefNum + ")");
    Gtype = -1; // undefined
    Gcolor = -1;
    Gtop = 0;
    Gleft = 0;
    Gtall = 0;
    Gwide = 0;
    Ginfo = 0;
    Parent = 0;
    Gdata = null;}} //~GameWgt

// Index to Selected Useful GameWgt Methods (cuz BlueJ has none)..
//   // Graphics Engine Data..
//   // Handy (private/utility) methods..
//   WhoCalled  CharAt  AddPair
//   // Widget public methods..
//   SetColor  AddPrntsPos  IsVisible  AddElemt  CheckIt
//   DrawMe  SetPosn  RadioBtn

  // (javagame) (WG)
